/*
 * Minidxr2 - Linux DVD player. Mostly (C) Jack Whitham, 2001
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>


#include "cmdline-control.h"
#include "control.h"
#include "parser.h"
#include "buffer.h"
#include "device.h"
#include "player.h"
#include "ifo_parse.h"
#include "read_config.h"
#include "player.h"
#include "message.h"
#include "output.h"

#ifdef HAVE_READLINE
#include <readline/history.h>
#include <readline/readline.h>
#endif

#define CMD_LOAD_FILE					4
#define CMD_LOAD_TITLE					5
#define CMD_PLAY_POS					6
#define CMD_PLAY_CHAPTER				7
#define CMD_PLAY_TIME					8
#define CMD_PRINT_CURRENT_TIME_CODE		9
#define CMD_PRINT_CURRENT_CHAPTER		10
#define CMD_PRINT_CURRENT_POS			11
#define CMD_PAUSE						12
#define CMD_FORWARD						13
#define CMD_REWIND						14
#define CMD_UNPAUSE						15
#define CMD_PLAY						16
#define CMD_FAST_FORWARD				17
#define CMD_SET_ANGLE					18
#define CMD_SET_TRACK					19
#define CMD_SET_SUB						20
#define CMD_PRINT_CHAPTER_INFO			21
#define CMD_PRINT_LENGTH				22
#define CMD_PRINT_SUB_INFO				23
#define CMD_PRINT_AUDIO_INFO			24
#define CMD_STOP						25
#define CMD_UNLOAD						26
#define CMD_QUIT						27
#define CMD_SET_SUB_OFF					28
#define CMD_SET_DEFAULT_AUDIO_FORMAT	29
#define CMD_SET_DEFAULT_AUDIO_WIDTH		30
#define CMD_SET_DEFAULT_AUDIO_FREQ		31
#define CMD_LOAD_DVD					32
#define CMD_EJECT_DVD					33
#define CMD_PRINT_TOTAL_TIME			34
#define CMD_SET_VIDEO_FORMAT			35
#define CMD_HELP						36
#define CMD_PRINT_BITRATE				37
#define CMD_PRINT_KNOWN_LANGUAGES		38
#define CMD_SELECT_BEST_AUDIO			39
#define CMD_SELECT_BEST_SUB				40
#define CMD_FULLSCREEN					41
#define CMD_WINDOW						42
#define CMD_NO_OVERLAY					43


#ifdef HAVE_READLINE
static char * command_generator ( char * text, int state );
char ** cmdline_completion ( char * text, int start, int end );
#endif

static parser_table command_table[] = { 
		{ "load_files" , (void *) CMD_LOAD_FILE , NULL , STRING } ,
		{ "load_file" , (void *) CMD_LOAD_FILE , NULL , STRING } ,
		{ "load_title" , (void *) CMD_LOAD_TITLE , NULL , INTEGER } ,
		{ "play_pos" , (void *) CMD_PLAY_POS , NULL , INTEGER } ,
		{ "play_chapter" , (void *) CMD_PLAY_CHAPTER , NULL , INTEGER } ,
		{ "play_time" , (void *) CMD_PLAY_TIME , NULL , STRING } ,
		{ "pct" , (void *) CMD_PRINT_CURRENT_TIME_CODE , NULL , NONE } ,
		{ "pcc" , (void *) CMD_PRINT_CURRENT_CHAPTER , NULL , NONE } ,
		{ "pcp" , (void *) CMD_PRINT_CURRENT_POS , NULL , NONE } ,
		{ "pause" , (void *) CMD_PAUSE , NULL , NONE } ,
		{ "fwd" , (void *) CMD_FORWARD , NULL , NONE } ,
		{ "rwd" , (void *) CMD_REWIND , NULL , NONE } ,
		{ "unpause" , (void *) CMD_UNPAUSE , NULL , NONE } ,
		{ "play" , (void *) CMD_PLAY , NULL , NONE } ,
		{ "ffwd" , (void *) CMD_FAST_FORWARD , NULL , NONE } ,
		{ "set_angle" , (void *) CMD_SET_ANGLE , NULL , INTEGER } ,
		{ "set_track" , (void *) CMD_SET_TRACK , NULL , INTEGER } ,
		{ "set_def_audio_format" , (void *) CMD_SET_DEFAULT_AUDIO_FORMAT , NULL , STRING } ,
		{ "set_def_audio_width" , (void *) CMD_SET_DEFAULT_AUDIO_WIDTH , NULL , STRING } ,
		{ "set_def_audio_freq" , (void *) CMD_SET_DEFAULT_AUDIO_FREQ , NULL , STRING } ,
		{ "set_sub" , (void *) CMD_SET_SUB , NULL , INTEGER } ,
		{ "set_sub_off" , (void *) CMD_SET_SUB_OFF , NULL , NONE } ,
		{ "print_chapter_info" , (void *) CMD_PRINT_CHAPTER_INFO , NULL , NONE } ,
		{ "print_length" , (void *) CMD_PRINT_LENGTH , NULL , NONE } ,
		{ "print_sub_info" , (void *) CMD_PRINT_SUB_INFO , NULL , NONE } ,
		{ "print_audio_info" , (void *) CMD_PRINT_AUDIO_INFO , NULL , NONE } ,
		{ "print_total_time" , (void *) CMD_PRINT_TOTAL_TIME , NULL , NONE } ,
		{ "set_video_format" , (void *) CMD_SET_VIDEO_FORMAT , NULL , NONE } ,
		{ "stop" , (void *) CMD_STOP , NULL , NONE } ,
		{ "unload" , (void *) CMD_UNLOAD , NULL , NONE } ,
		{ "quit" , (void *) CMD_QUIT , NULL , NONE } ,
		{ "load_dvd" , (void *) CMD_LOAD_DVD , NULL , NONE } ,
		{ "eject" , (void *) CMD_EJECT_DVD , NULL , NONE } ,
		{ "help" , (void *) CMD_HELP , NULL , NONE } ,
		{ "print_bitrate" , (void *) CMD_PRINT_BITRATE , NULL , NONE } ,
		{ "print_known_langs" , (void *) CMD_PRINT_KNOWN_LANGUAGES , NULL , NONE } ,
		{ "set_best_track" , (void *) CMD_SELECT_BEST_AUDIO , NULL , STRING } ,
		{ "set_best_sub" , (void *) CMD_SELECT_BEST_SUB , NULL , STRING } ,
		{ "fullscreen" , (void *) CMD_FULLSCREEN , NULL , NONE } ,
		{ "window" , (void *) CMD_WINDOW , NULL , NONE } ,
		{ "no_overlay" , (void *) CMD_NO_OVERLAY , NULL , NONE } ,
		{ "" , NULL , NULL , 0 } } ;

void * cmdline_control_procedure ( void * arg )
{
	player_info_t* player_info = (player_info_t*) arg;
#ifdef HAVE_READLINE
	char			* command ;
#else
	char			command [ BUFFER_LENGTH ] ;
#endif
	char			temp_string [ BUFFER_LENGTH ] ;
	int				temp_int ;
	char			* second_word ;
	char			* endptr ;
	char			* err ;
	char			* err2 ;
	char			errbuf [ BUFFER_LENGTH ] ;
	int				command_num , command_param , param_int = 0 ;
	int				hour , min , sec ;
	
	printf ( "%s: command line interface\n" , ANNOUNCE ) ;
	printf ( "Type 'help' for command listing\n\n" ) ;
	
#ifdef HAVE_READLINE
	using_history ();
	rl_readline_name = ANNOUNCE ;
	rl_attempted_completion_function = ( CPPFunction * ) cmdline_completion;
#endif

	player_info -> h_state = DVD_UNLOADED ;
	err = dvd_initialise_overlay ( player_info , 
					& player_info -> default_colour_key ) ;
	if ( err == NULL )
	{
		err = dvd_set_view_mode ( player_info , OUTPUT_VIEW_WINDOW ) ;
	} else {
		puts ( err ) ;
	}

	if ( player_info -> immediate_playback == IP_FILES )
	{
		errbuf [ 0 ] = '\0' ;
		if ( dvd_open_file ( player_info , 
				player_info -> immediate_playback_names ,
				player_info -> immediate_playback_files , 
				errbuf , BUFFER_LENGTH_1 ) )
		{
			(void) dvd_play ( player_info ) ;
		} else {
			printf ( "Immediate file play: %s\n" , errbuf ) ;
		}
	} else if ( player_info -> immediate_playback == IP_DVD )
	{
		err = device_mount ( player_info ) ;	/* ignore mount errors */
		if ( dvd_open_title ( player_info , -1 , errbuf , BUFFER_LENGTH_1 ) )
		{
			(void) dvd_play ( player_info ) ;
		} else {
			printf ( "Immediate dvd play: %s\n" , errbuf ) ;
		}
	}

	while ( player_info -> h_state != DVD_QUIT )
	{
#ifdef HAVE_READLINE
		if ( (command = readline (">> ")) == NULL) {
			printf ( "\n" );
			err = dvd_quit ( player_info ) ;
			if ( err != NULL ) {
				puts ( err );
			}
			return ( NULL );
		}
		add_history ( command );
#else
		printf ( ">> " ) ;
		fgets ( command , BUFFER_LENGTH_1 , stdin ) ;
#endif
		
		remove_whitespace ( command ) ;

		second_word = index ( command , ' ' ) ;
		if ( second_word == NULL )
		{
			second_word = "" ;
		} else {
			second_word [ 0 ] = '\0' ;
			second_word = & second_word [ 1 ] ;
			remove_whitespace ( second_word ) ;
		}

		if ( strlen ( command ) > 0 )
		{
			err = NULL ;

			command_param = parse_type ( command , command_table ,
					(void **) & command_num , NULL , "command" , errbuf ) ;

			if ( command_param == INTEGER )
			{
				param_int = (int) strtol ( second_word , & endptr , 0 ) ;
				if ( endptr == second_word )
				{
					puts ( "That command takes an integer parameter." ) ;
					continue ;
				}
			} else if ( command_param == STRING )
			{
				if ( strlen ( second_word ) < 1 )
				{
					puts ( "That command takes a string parameter." ) ;
					continue ;
				}
			} else if ( command_param != NONE )
			{
				puts ( errbuf ) ;
				continue ;
			}

			switch ( command_num )
			{
				case CMD_LOAD_FILE :
						{
							int		n , i ;
							int		rc ;
							char	* flist [ MAX_FILES ] ;

							n = parse_file_list ( second_word , flist , MAX_FILES ) ;
							if ( n > 0 )
							{
								rc = dvd_open_file ( player_info , flist , n , errbuf , BUFFER_LENGTH_1 ) ;
								for ( i = 0 ; i < n ; i ++ )
								{
									free ( flist [ i ] ) ;
								}
								if ( ! rc )
								{
									err = errbuf ;
								}
							} else {
								err = "Usage: load_files <file list....>" ;
							}
						}
						break ;
				case CMD_LOAD_TITLE :
						if ( ! dvd_open_title ( player_info , param_int , errbuf , BUFFER_LENGTH_1 ) )
						{
							err = errbuf ;
						}
						break ;
				case CMD_PLAY_POS :
						err = dvd_seek_position ( player_info , param_int ) ;
						break ;
				case CMD_PLAY_CHAPTER :
						err = dvd_seek_chapter ( player_info , param_int ) ;
						break ;
				case CMD_PLAY_TIME :
						hour = min = sec = 0 ;
						if ( sscanf ( second_word , "%d %d %d" ,
								& hour , & min , & sec ) != 3 )
						{
							err = "Usage: play_time <hour> <min> <sec>" ;
						} else {
							err = dvd_seek_time ( player_info , hour , min , sec ) ;
						}
						break ;
				case CMD_PRINT_CURRENT_TIME_CODE :
						err = dvd_get_time_code ( player_info , temp_string , 79 ) ;
						if ( err == NULL )
						{
							err = temp_string ;
						}
						break ;
				case CMD_PRINT_CURRENT_CHAPTER :
						err = dvd_get_chapter ( player_info , & temp_int ) ;
						if ( err == NULL )
						{
							sprintf ( temp_string , "%d\n" , temp_int ) ;
							err = temp_string ;
						}
						break ;
				case CMD_PRINT_CURRENT_POS :
						err = dvd_get_position ( player_info , & temp_int ) ;
						if ( err == NULL )
						{
							sprintf ( temp_string , "%d\n" , temp_int ) ;
							err = temp_string ;
						}
						break ;
				case CMD_PAUSE :
						err = dvd_pause ( player_info ) ;
						break ;
				case CMD_FORWARD :
						if (( err = dvd_get_position ( player_info , & temp_int )) == NULL )
						{
							temp_int += 
									( player_info -> h_state != DVD_FAST_FORWARD ) ?
										JUMP_DISTANCE : FFWD_JUMP_DISTANCE ;
							err = dvd_seek_position ( player_info , temp_int ) ;
						}
						break ;
				case CMD_REWIND :
						if (( err = dvd_get_position ( player_info , & temp_int )) == NULL )
						{
							temp_int -= 
									( player_info -> h_state != DVD_FAST_FORWARD ) ?
										JUMP_DISTANCE : FFWD_JUMP_DISTANCE ;
							if ( temp_int < 0 )
							{
								temp_int = 0 ;
							}
							err = dvd_seek_position ( player_info , temp_int ) ;
						}
						break ;
				case CMD_UNPAUSE :
				case CMD_PLAY :
						err = dvd_play ( player_info ) ;
						break ;
				case CMD_FAST_FORWARD :
						err = dvd_ffwd ( player_info ) ;
						break ;
				case CMD_SET_ANGLE :
						err = dvd_set_angle ( player_info , param_int ) ;
						break ;
				case CMD_SET_TRACK :
						err = dvd_set_track ( player_info , param_int ) ;
						break ;
				case CMD_SET_DEFAULT_AUDIO_FORMAT :
						if ( parse_type ( second_word , audio_stream_table ,
							(void **) & temp_int , NULL , "audio format" , 
							temp_string ) == INTEGER )
						{
							dvd_set_default_audio_format ( player_info , temp_int ) ;
							err = NULL ;
						} else {
							err = temp_string ;
						}
						break ;
				case CMD_SET_DEFAULT_AUDIO_WIDTH :
						if ( parse_type ( second_word , audio_width_table ,
							(void **) & temp_int , NULL , "audio width" , 
							temp_string ) == INTEGER )
						{
							dvd_set_default_audio_width ( player_info , temp_int ) ;
							err = NULL ;
						} else {
							err = temp_string ;
						}
						break ;
				case CMD_SET_DEFAULT_AUDIO_FREQ :
						if ( parse_type ( second_word , audio_freq_table ,
							(void **) & temp_int , NULL , "audio frequency" , 
							temp_string ) == INTEGER )
						{
							dvd_set_default_audio_freq ( player_info , temp_int ) ;
							err = NULL ;
						} else {
							err = temp_string ;
						}
						break ;
				case CMD_SET_SUB :
						err = dvd_set_sub ( player_info , param_int ) ;
						break ;
				case CMD_SET_SUB_OFF :
						err = dvd_set_sub ( player_info , -1 ) ;
						break ;
				case CMD_SET_VIDEO_FORMAT :
						if ( parse_type ( second_word , video_format_table ,
							(void **) & temp_int , NULL , "video format" , 
							temp_string ) == INTEGER )
						{
							pthread_mutex_lock ( & ( player_info -> audio_mutex ) ) ;
							player_info -> default_video_format = temp_int ;
							output_set_scaling_mode ( player_info -> out_plugin , temp_int ) ;
							pthread_mutex_unlock ( & ( player_info -> audio_mutex ) ) ;
							err = NULL ;
						} else {
							err = temp_string ;
						}
						break ;
						break ;
				case CMD_PRINT_CHAPTER_INFO :
						if (( err = dvd_print_chapter_info ( player_info , temp_string , BUFFER_LENGTH_1 )) == NULL )
						{
							err = temp_string ;
						}
						break ;
				case CMD_PRINT_SUB_INFO :
						if (( err = dvd_print_sub_info ( player_info , temp_string , BUFFER_LENGTH_1 )) == NULL )
						{
							err = temp_string ;
						}
						break ;
				case CMD_PRINT_AUDIO_INFO :
						if (( err = dvd_print_audio_info ( player_info , temp_string , BUFFER_LENGTH_1 )) == NULL )
						{
							err = temp_string ;
						}
						break ;
				case CMD_PRINT_LENGTH :
						err = dvd_get_length ( player_info , & temp_int ) ;
						if ( err == NULL )
						{
							sprintf ( temp_string , "%d\n" , temp_int ) ;
							err = temp_string ;
						}
						break ;
				case CMD_PRINT_TOTAL_TIME :
						if (( err = dvd_print_total_time ( player_info , temp_string , BUFFER_LENGTH_1 )) == NULL )
						{
							err = temp_string ;
						}
						break ;
				case CMD_SELECT_BEST_AUDIO :
						temp_int = dvd_find_best_audio_track ( player_info , second_word ) ;
						if ( temp_int > 0 )
						{
							err = dvd_set_track ( player_info , temp_int ) ;
						} else {
							err = "Unable to determine best audio track.\n" ;
						}
						break ;
				case CMD_SELECT_BEST_SUB :
						temp_int = dvd_find_best_sub_track ( player_info , second_word ) ;
						if ( temp_int > 0 )
						{
							err = dvd_set_sub ( player_info , temp_int ) ;
						} else {
							err = "Unable to determine best subtitle track.\n" ;
						}
						break ;
				case CMD_STOP :
						err = dvd_stop ( player_info ) ;
						break ;
				case CMD_UNLOAD :
						err = dvd_unload ( player_info ) ;
						break ;
				case CMD_QUIT :
						err = dvd_quit ( player_info ) ;
						break ;
				case CMD_LOAD_DVD :
						err = device_mount ( player_info ) ;	/* ignore mount errors */
						if ( ! dvd_open_title ( player_info , -1 , errbuf , BUFFER_LENGTH_1 ) )
						{
							err = errbuf ;
						} else {
							err = NULL ;
						}
						break ;
				case CMD_EJECT_DVD :
						err = dvd_unload ( player_info ) ;
						if ( err == NULL )
						{
							err2 = device_unmount ( player_info ) ;
							err = device_eject ( player_info ) ;
							if ( err != NULL )
							{
								err = err2 ; 
								/* unmount errors take priority over eject errors */
							}
						}
						break ;
				case CMD_PRINT_BITRATE :
						temp_int = output_get_bitrate ( player_info -> out_plugin ) ;
						sprintf ( temp_string , "%d\n" , temp_int ) ;
						err = temp_string ;
						break ;
				case CMD_PRINT_KNOWN_LANGUAGES :
						printf ( "libifo known language codes:\n" ) ;
						err = dvd_list_languages () ;
						SAFE_STRNCPY ( temp_string , err , BUFFER_LENGTH_1 ) ;
						free ( err ) ;
						err = temp_string ;
						break ;
				case CMD_FULLSCREEN :
						err = dvd_set_view_mode ( player_info , OUTPUT_VIEW_FULLSCREEN ) ;
						break ;
				case CMD_WINDOW :
						err = dvd_set_view_mode ( player_info , OUTPUT_VIEW_WINDOW ) ;
						break ;
				case CMD_NO_OVERLAY :
						err = dvd_set_view_mode ( player_info , OUTPUT_VIEW_OFF ) ;
						break ;
				case CMD_HELP :
						printf ( "Commands are:\n" ) ;
						print_all_keys ( command_table ) ;
						printf ( "\n" ) ;
						err = NULL ;
						break ;
				default :
						err = "Invalid command, use 'help' to list all commands.\n" ;
						break ;
			}

			if ( err != NULL )
			{
				puts ( err ) ;
			}
		}
#ifdef HAVE_READLINE
		free ( command );
#endif
		temp_int = 0 ;
		err = NULL ;
		while ( dvd_get_message ( player_info , & temp_int , & err ) )
		{
			switch ( temp_int )
			{
				case MSG_CSS_ERROR :
						printf ( "CSS error: %s\n" , err ) ;
						err = dvd_unload ( player_info ) ;
						break ;
				case MSG_WRITE_ERROR :
						printf ( "Error writing to output: %s\n" , err ) ;
						err = dvd_unload ( player_info ) ;
						break ;
				case MSG_IFO_ERROR :
						printf ( "IFO parsing error: %s\n" , err ) ;
						/* Non-fatal. Don't unload disc */
						break ;
				case MSG_END_OF_FILE :
						err = dvd_stop ( player_info ) ;
						break ;
			}
			if ( err != NULL )
			{
				puts ( err ) ;
			}
			err = NULL ;
		}
	}
	return NULL ;
}

#ifdef HAVE_READLINE

char ** cmdline_completion ( char * text, int start, int end )
{
	if ( start == 0 ) 
	{
		return ( completion_matches ( text, command_generator ) );
	} else {
		return NULL;
	}
}

static char * command_generator ( char * text, int state )
{
	int index , matches , len ;
	char * match_text = NULL ;

	/* This implementation is slightly less efficient than the readline
	 * example and the minidxr2 readline patch by Ethan Blanton. I 
	 * rewrote it like this because the original used a couple of globals, and
	 * minidxr2 is global free on principle :). */


	len = strlen ( text ) ;
	/* first pass scan through the table to see how many matches exist */
	index = 0 ;
	matches = 0 ;
	while ( strlen ( command_table [ index ] . key ) > 0 )
	{
		if ( strncasecmp ( command_table [ index ] . key , text , len ) == 0 )
		{
			match_text = command_table [ index ] . key ;
			matches ++ ;
		}
		index ++ ;
	}

	if ( state >= matches )
	{
		return NULL ;	/* no more matches */
	}

	if ( matches == 0 ) /* no matches at all */
	{
		return NULL ;
	} else if ( matches == 1 )
	{
		return strdup ( match_text ) ;
	} else {
		int		skip = state ;

		/* second pass - find correct match for state */
		index = 0 ;
		while ( strlen ( command_table [ index ] . key ) > 0 )
		{
			if ( strncasecmp ( command_table [ index ] . key , text, len ) == 0 )
			{
				skip -- ;
				if ( skip < 0 )
				{
					return strdup ( command_table [ index ] . key ) ;
				}
			}
			index ++ ;
		}
		return NULL ;
	}
}


#endif

